#                             pyGoDaddyAUpdater
#                  Copyright (C) 2019 - Javinator9889
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#                   (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#               GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
from ..logging_utils.utils import LoggingHandler
from ..logging_utils.utils import cleanup_old_logs
from ..logging_utils.utils import setup_logging



#                             pyGoDaddyAUpdater
#                  Copyright (C) 2019 - Javinator9889
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#                   (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#               GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging


def cleanup_old_logs(log_file: str):
    import tarfile
    import os

    tar_log_filename = log_file + ".tar.gz"

    if os.path.exists(log_file):
        if os.path.exists(tar_log_filename):
            os.remove(tar_log_filename)
        with tarfile.open(tar_log_filename, "w:gz") as tar:
            tar.add(log_file, arcname=os.path.basename(log_file))
            tar.close()
            os.remove(log_file)


def setup_logging(logger_name: str, log_file: str, level=logging.DEBUG,
                  formatter: str = "%(process)d - %(asctime)s | [%(levelname)s]: %(message)s"):
    from os import path
    from os import makedirs

    log_dir = path.dirname(path.abspath(log_file))
    if not path.exists(log_dir):
        makedirs(log_dir)

    cleanup_old_logs(log_file)
    new_logging = logging.getLogger(logger_name)
    logging_formatter = logging.Formatter(formatter)
    logging_file_handler = logging.FileHandler(log_file, mode="w")

    logging_file_handler.setFormatter(logging_formatter)

    new_logging.setLevel(level)
    new_logging.addHandler(logging_file_handler)

    return logging_file_handler


class LoggingHandler(object):
    class __LoggingHandler(object):
        def __init__(self, logs: list):
            self.__logs = logs

        def debug(self, msg):
            for log in self.__logs:
                log.debug(msg)

        def info(self, msg):
            for log in self.__logs:
                log.info(msg)

        def error(self, msg):
            for log in self.__logs:
                log.error(msg)

        def warning(self, msg):
            for log in self.__logs:
                log.warning(msg)

        def critical(self, msg):
            for log in self.__logs:
                log.critical(msg)

        def exception(self, msg, *args, exc_info: bool = True, **kwargs):
            for log in self.__logs:
                log.exception(msg, args, exc_info, kwargs)

        def get_loggers(self) -> list:
            return self.__logs

    __instance = None

    def __new__(cls, *args, **kwargs):
        if not LoggingHandler.__instance:
            logs = kwargs.get("logs")
            if not logs or len(logs) == 0:
                raise AttributeError("At least kwarg \"log\" (a list of the loggers) must be provided")
            LoggingHandler.__instance = LoggingHandler.__LoggingHandler(logs)
        return LoggingHandler.__instance

    def __getattr__(self, item):
        return getattr(self.__instance, item)

    def __setattr__(self, key, value):
        return setattr(self.__instance, key, value)

    def debug(self, msg):
        self.__instance.debug(msg)

    def info(self, msg):
        self.__instance.info(msg)

    def error(self, msg):
        self.__instance.error(msg)

    def warning(self, msg):
        self.__instance.warning(msg)

    def critical(self, msg):
        self.__instance.critical(msg)

    def exception(self, msg, *args, exc_info: bool = True, **kwargs):
        self.__instance.exception(msg, args, exc_info, kwargs)

    def get_loggers(self) -> list:
        return self.__instance.get_loggers()



#                             pyGoDaddyAUpdater
#                  Copyright (C) 2019 - Javinator9889
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#                   (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#               GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.


class UserPreferences(object):
    def __init__(self, **kwargs):
        if kwargs:
            try:
                self.__domain = kwargs["domain"]
                self.__name = kwargs["name"]
                self.__time = kwargs["time"]
                self.__key = kwargs["key"]
                self.__secret = kwargs["secret"]
                self.__pid = kwargs["pid"]
                self.__log = kwargs["log"]
            except KeyError:
                raise AttributeError("Some value was not provided while creating user preferences\n"
                                     "Values are:\n"
                                     "  - Domain\n"
                                     "  - Name (of A Record)\n"
                                     "  - Time (update time)\n"
                                     "  - Key\n"
                                     "  - Secret\n")
        else:
            self.__domain = None
            self.__name = None
            self.__time = None
            self.__key = None
            self.__secret = None
            self.__pid = None
            self.__log = None
        self.__latest_ip = "0.0.0.0"
        self.__daemonize = True

    @staticmethod
    def get_preferences_file():
        import os

        return os.path.dirname(os.path.abspath(__file__)) + "/user.preferences"

    def load_preferences(self):
        import pickle
        import os

        from base64 import b64decode

        if os.path.exists("user.preferences"):
            with open("user.preferences", "rb") as fpreferences:
                preferences = pickle.load(fpreferences)
            self.__domain = preferences["domain"]
            self.__secret = b64decode(preferences["secret"]).decode("utf-8")
            self.__time = preferences["time"]
            self.__key = b64decode(preferences["key"]).decode("utf-8")
            self.__name = preferences["name"]
            self.__latest_ip = preferences["latest_ip"]
            self.__pid = preferences["pid"]
            self.__log = preferences["log"]
        else:
            raise FileNotFoundError("There are no saved user preferences. Call \"save_preferences\" the first time")

    def save_preferences(self, filename="user.preferences"):
        import pickle

        from base64 import b64encode
        from os import path
        from os import makedirs

        preferences = {"domain": self.__domain,
                       "name": self.__name,
                       "time": self.__time,
                       "key": b64encode(bytes(self.__key, "utf-8")),
                       "secret": b64encode(bytes(self.__secret, "utf-8")),
                       "latest_ip": self.__latest_ip,
                       "pid": self.__pid,
                       "log": self.__log}
        file_dir = path.dirname(path.abspath(filename))
        if not path.exists(file_dir):
            makedirs(path=file_dir, exist_ok=True)
        with open(filename, "wb") as fpreferences:
            pickle.dump(preferences, fpreferences, pickle.HIGHEST_PROTOCOL)

    def get_domain(self):
        return self.__domain

    def get_name(self):
        return self.__name

    def get_time(self):
        return self.__time

    def get_key(self):
        return self.__key

    def get_secret(self):
        return self.__secret

    def get_latest_ip(self):
        return self.__latest_ip

    def get_pid_file(self):
        return self.__pid

    def get_log_file(self):
        return self.__log

    def is_running_as_daemon(self):
        return self.__daemonize

    def set_domain(self, domain):
        self.__domain = domain

    def set_name(self, name):
        self.__name = name

    def set_time(self, time):
        self.__time = time

    def set_key(self, key):
        self.__key = key

    def set_secret(self, secret):
        self.__secret = secret

    def set_latest_ip(self, ip):
        self.__latest_ip = ip

    def set_pid_file(self, pid):
        self.__pid = pid

    def set_log_file(self, log):
        self.__log = log

    def run_as_daemon(self, daemonize: bool):
        self.__daemonize = daemonize

    @staticmethod
    def are_preferences_stored():
        import os

        return os.path.exists("user.preferences")



#                             pyGoDaddyAUpdater
#                  Copyright (C) 2019 - Javinator9889
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#                   (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#               GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.


def get_machine_public_ip():
    import urllib.request

    return urllib.request.urlopen('https://ident.me').read().decode('utf8')


class GoDaddy(object):
    def __init__(self, domain, name, key, secret):
        self.__domain = domain
        self.__name = name
        self.__headers = "sso-key {0}:{1}".format(key, secret)

    def get_godaddy_latest_ip(self):
        from urllib.request import Request, urlopen
        from re import search

        request = Request("https://api.godaddy.com/v1/domains/{0}/records/A/{1}".format(self.__domain, self.__name))
        request.add_header("Authorization", self.__headers)
        result = urlopen(request).read().decode('utf8')

        ip = search("([0-9]{1,3}\.){3}[0-9]{1,3}", result)
        return ip.group(0) if ip else "0.0.0.0"

    def set_goddady_ip(self, ip):
        from urllib.request import Request, urlopen
        try:
            from ujson import dumps
        except ImportError:
            from json import dumps

        data = dumps([{"data": ip, "ttl": 600, "name": self.__name, "type": "A"}])
        request = Request(url="https://api.godaddy.com/v1/domains/{0}/records/A/{1}".format(self.__domain, self.__name),
                          data=data.encode("utf-8"),
                          headers={"Authorization": self.__headers, "Content-Type": "application/json"},
                          method="PUT")
        return urlopen(request).getcode()



#                             pyGoDaddyAUpdater
#                  Copyright (C) 2019 - Javinator9889
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#                   (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#               GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
from argparse import ArgumentParser
from argparse import SUPPRESS
from logging import getLogger
from time import sleep
from os import path
from os import makedirs

from daemonize import Daemonize

from .logging_utils import LoggingHandler
from .logging_utils import setup_logging
from .network import GoDaddy
from .network import get_machine_public_ip
from .preferences import UserPreferences
from .values import description

preferences = UserPreferences()


def main():
    loop_continuation = True
    log = LoggingHandler(logs=[getLogger("appLogger")])
    net = GoDaddy(preferences.get_domain(), preferences.get_name(), preferences.get_key(), preferences.get_secret())
    try:
        while loop_continuation:
            current_ip = get_machine_public_ip()
            log.info("Current machine IP: \"{0}\"".format(current_ip))
            if preferences.get_latest_ip() == "0.0.0.0":
                preferences.set_latest_ip(net.get_godaddy_latest_ip())
                log.warning("User saved latest IP is not up to date - downloading GoDaddy A Record value: \"{0}\""
                            .format(preferences.get_latest_ip()))
            if preferences.get_latest_ip() != current_ip:
                log.info("IP needs an upgrade - OLD IP: {0} | NEW IP: {1}"
                         .format(preferences.get_latest_ip(), current_ip))
                result = net.set_goddady_ip(current_ip)
                log.info("IP updated correctly! - Operation return code: {0}".format(result))
                log.debug("Updating saved IP...")
                preferences.set_latest_ip(current_ip)
            else:
                log.info("IP has not changed - skipping")
            if not preferences.is_running_as_daemon():
                log.info("This script is only executed once. Finishing...")
                loop_continuation = False
            else:
                log.info("Next check in about {0} minute{1}"
                         .format((preferences.get_time() / 60),
                                 's' if (preferences.get_time() / 60) > 1 else ''))
                sleep(preferences.get_time())
    except KeyboardInterrupt:
        log.warning("Received SIGINT - exiting...")
        preferences.save_preferences()
        exit(1)


def parser():
    is_first_execution = not preferences.are_preferences_stored()
    args = ArgumentParser(description=description,
                          allow_abbrev=False)
    args.add_argument("--domain",
                      type=str,
                      required=is_first_execution,
                      help="GoDaddy domain to be updated.")
    args.add_argument("--name",
                      type=str,
                      required=is_first_execution,
                      help="GoDaddy 'A' Record name.")
    args.add_argument("--time",
                      type=int,
                      default=SUPPRESS,
                      required=is_first_execution,
                      help="Time (in minutes) to check for updated IP (defaults: 5 min.) - must be higher than 0.")
    args.add_argument("--key",
                      type=str,
                      required=is_first_execution,
                      help="GoDaddy developer key.")
    args.add_argument("--secret",
                      type=str,
                      required=is_first_execution,
                      help="GoDaddy developer secret.")
    args.add_argument("--no_daemonize",
                      action="store_true",
                      required=False,
                      default=False,
                      help="By default, the program runs as a daemon in background. With this option enabled, "
                           "the program will run only once and then exit.")
    args.add_argument("--pid",
                      type=str,
                      default=SUPPRESS,
                      required=False,
                      metavar="PID FILE",
                      help="Specifies a custom PID file for storing current daemon PID.")
    args.add_argument("--log",
                      type=str,
                      default=SUPPRESS,
                      required=False,
                      metavar="LOG FILE",
                      help="Specifies a custom LOG file for storing current daemon logs.")
    args.add_argument("--preferences",
                      type=str,
                      default="user.preferences",
                      required=False,
                      metavar="PREFERENCES FILE",
                      help="Provide a custom preferences file - useful for multiple running daemon for different 'A'"
                           "Records. NOTICE THAT YOU MUST PROVIDE ALL REQUIRED PARAMS FOR A NEW CONFIGURATION")
    args.add_argument("--user",
                      type=str,
                      default=None,
                      required=False,
                      metavar="USERNAME",
                      help="Run the daemon as the specified user.")
    args.add_argument("--group",
                      type=str,
                      default=None,
                      required=False,
                      metavar="GROUP NAME",
                      help="Run the daemon as the specified group.")
    p_args = args.parse_args()
    should_save_preferences = False
    if p_args.domain:
        preferences.set_domain(p_args.domain)
        should_save_preferences = True
    if p_args.name:
        preferences.set_name(p_args.name)
        should_save_preferences = True
    if "time" in p_args:
        preferences.set_time(p_args.time * 60)
        should_save_preferences = True
    else:
        if preferences.get_time() is None:
            preferences.set_time(300)
    if p_args.key:
        preferences.set_key(p_args.key)
        should_save_preferences = True
    if p_args.secret:
        preferences.set_secret(p_args.secret)
        should_save_preferences = True
    if p_args.no_daemonize:
        preferences.run_as_daemon(not p_args.no_daemonize)
    if "pid" in p_args:
        preferences.set_pid_file(p_args.pid)
        should_save_preferences = True
    else:
        if preferences.get_pid_file() is None:
            preferences.set_pid_file("/var/run/pygoddady.pid")
    if "log" in p_args:
        preferences.set_log_file(p_args.log)
        should_save_preferences = True
    else:
        if preferences.get_log_file() is None:
            preferences.set_log_file("/var/log/pygoddady.log")
    user = p_args.user
    group = p_args.group

    if preferences:
        if not (p_args.domain and p_args.name and p_args.key and p_args.secret):
            print("You must provide the required params for a new preferences file")
    if should_save_preferences:
        preferences.save_preferences(p_args.preferences)
    if not is_first_execution:
        preferences.load_preferences()
    file_handler = setup_logging("appLogger", preferences.get_log_file())
    fds = [file_handler.stream.fileno()]
    pid_dir = path.dirname(path.abspath(preferences.get_pid_file()))
    if not path.exists(pid_dir):
        makedirs(path=pid_dir, exist_ok=True)

    daemon = Daemonize(app="pyGoDaddyDaemon",
                       pid=preferences.get_pid_file(),
                       action=main,
                       keep_fds=fds,
                       user=user,
                       group=group,
                       logger=getLogger("appLogger"))
    daemon.start()


if __name__ == '__main__':
    parser()



from sys import version
from os import path
from setuptools import setup

if version < '3':
    raise RuntimeError("Python 3 is, at least, needed")

this = path.abspath(path.dirname(__file__))
with open(path.join(this, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='pyGoDaddyUpdater',
    version='1.01',
    packages=['pyGoDaddyUpdater',
              'pyGoDaddyUpdater.values',
              'pyGoDaddyUpdater.network',
              'pyGoDaddyUpdater.preferences',
              'pyGoDaddyUpdater.logging_utils'],
    url='https://gitlab.javinator9889.com/Javinator9889/pyGoDaddyAUpdater',
    license='GPL-3.0',
    author='Javinator9889',
    author_email='javialonso007@hotmail.es',
    description='DDNS service for dynamically update GoDaddy A Records',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=False,
    zip_safe=True,
    download_url="https://gitlab.javinator9889.com/Javinator9889/pyGoDaddyAUpdater/repository/master/archive.zip",
    entry_points={
        'console_scripts': ['godaddy_ddns=pyGoDaddyUpdater.__main__:parser']
    },
    install_requires=['daemonize'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)
